"""Testing MongoDB data sources for the shleem python package."""

import pytest
from strct.general import stable_hash_builtins_strct

import shleem


def test_mongo_sources():

    # defining a MongoDB server source
    shleem_test_server = shleem.mongodb.server("shleem_test_server")
    assert repr(shleem_test_server) == (
        "MongoDB server DataSource: shleem_test_server")
    assert shleem_test_server.identifier == 'shleem_test_server'
    assert shleem_test_server.server_name == 'shleem_test_server'

    # db source
    test_db = shleem_test_server['shleem_test']
    test_db2 = shleem_test_server.shleem_test
    test_db3 = shleem_test_server.db('shleem_test')
    assert test_db == test_db2
    assert test_db2 == test_db3
    assert repr(test_db) == (
        "MongoDB database DataSource: shleem_test_server.shleem_test")
    assert test_db.identifier == 'shleem_test_server.shleem_test'
    assert test_db.db_name == 'shleem_test'

    # collection source
    examp = test_db['example_data_collection']
    examp2 = test_db.example_data_collection
    examp3 = test_db.collection('example_data_collection')
    assert examp == examp2
    assert examp == examp3
    assert repr(examp) == (
        "MongoDB collection DataSource: shleem_test_server.shleem_test"
        ".example_data_collection")
    assert examp.identifier == (
        'shleem_test_server.shleem_test.example_data_collection')
    assert examp.collection_name == 'example_data_collection'

    # query data tap
    query = {"borough": "Queens"}
    queens_people = examp.query(query)
    assert repr(queens_people) == (
        "MongoDB query DataSource: shleem_test_server.shleem_test"
        ".example_data_collection.{}".format(
            stable_hash_builtins_strct(query)))
    assert queens_people.identifier == (
        "shleem_test_server.shleem_test.example_data_collection.{}".format(
            stable_hash_builtins_strct(query)))
    assert queens_people.source_type == 'MongoDB'

    # checking the client
    client = shleem_test_server._get_connection()
    some_doc = client['shleem_test']['example_data_collection'].find_one()
    assert some_doc['restaurant_id'] == '30112340'

    # tapping the source
    cursor = queens_people.tap()
    other_doc = cursor.next()
    assert other_doc['restaurant_id'] == '40356068'

    def getter(field_name):
        return lambda **kwargs: kwargs[field_name]

    # parameterized query
    param_query_dict = {"address.zipcode": {
        "$gte": getter("min_val"), "$lte": getter("max_val")
    }}
    # first without identifier, so we check all the code!
    zipcode_range = examp.query(param_query_dict)
    zipcode_range = examp.query(
        query_dict=param_query_dict, identifier="zipcode_range")
    min_val = 11249
    max_val = 11300
    cursor = zipcode_range.tap(min_val=str(min_val), max_val=str(max_val))
    documents = [doc for doc in cursor]
    assert len(documents) == 163
    for doc in documents:
        zipcode = doc['address']['zipcode']
        assert int(zipcode) >= min_val and int(zipcode) <= max_val

    # with skip and limit
    some_limit = 17
    zipcode_range2 = examp.query(
        query_dict=param_query_dict, skip=10, limit=some_limit)
    cursor = zipcode_range2.tap(min_val=str(min_val), max_val=str(max_val))
    documents = [doc for doc in cursor]
    assert len(documents) == some_limit
    for doc in documents:
        zipcode = doc['address']['zipcode']
        assert int(zipcode) >= min_val and int(zipcode) <= max_val

    # aggregation data tap
    agg_pipeline = [
        {'$group': {'_id': '$borough', 'count': {'$sum': 1}}}
    ]
    borough_counts = examp.aggregation(agg_pipeline)
    exp_id = (
        'shleem_test_server.shleem_test.example_data_collection.'
        '7098918b9417525bcb7126c8978ad3c10ac4a352a2238d3e614cfcd50b6b82bc')
    assert borough_counts.identifier == exp_id
    exp_repr = "MongoDB aggregation DataSource: {}".format(exp_id)
    assert repr(borough_counts) == exp_repr

    # tap the source 2: the aggregation strikes back
    cursor = borough_counts.tap()
    res = {doc['_id']: doc['count'] for doc in cursor}
    assert res['Missing'] == 51
    assert res['Queens'] == 5656

    # parameterized aggregation
    param_agg = [
        {"$match": {"address.zipcode": {"$gte": "11695"}}},
        {"$unwind": "$grades"},
        {"$group": {
            "_id": {"name": "$name", "address": "$address"},
            "sum_score": {"$sum": "$grades.score"},
            "num_score": {"$sum": 1}
        }},
        {"$project": {
            "_id": 1, "avg_score": {"$divide": ["$sum_score", "$num_score"]}
        }},
        {"$match": {
            "avg_score": {
                "$gt": lambda **kwargs: kwargs["avg_score_threshold"]
            }
        }}
    ]
    # first without and identifier, so we check all the code handling that!
    by_avg_score_threshold = examp.aggregation(param_agg)
    # now with an identifier!
    by_avg_score_threshold = examp.aggregation(
        param_agg, "restaurant_by_avg_score_threshold")
    thresh = 13.5
    cursor = by_avg_score_threshold.tap(avg_score_threshold=thresh)
    for doc in cursor:
        assert doc['avg_score'] > thresh

    # parameterized aggregation with a callable in a list
    param_agg2 = [
        {"$match": {"address.zipcode": {"$gte": "11695"}}},
        {"$unwind": "$grades"},
        {"$group": {
            "_id": {"name": "$name", "address": "$address"},
            "sum_score": {"$sum": "$grades.score"}
        }},
        {"$project": {
            "_id": 1,
            "avg_score": {"$divide": ["$sum_score", getter("normalizer")]}
        }},
        {"$match": {
            "avg_score": {
                "$gt": lambda **kwargs: kwargs["avg_score_threshold"]
            }
        }}
    ]
    by_norm_score_threshold = examp.aggregation(param_agg2)
    thresh2 = 13.5
    cursor = by_norm_score_threshold.tap(
        avg_score_threshold=thresh2, normalizer=4)
    all_res = [doc for doc in cursor]
    assert len(all_res) == 2
    for doc in all_res:
        assert doc['avg_score'] > thresh2


def test_missing_server():
    missing_server = shleem.mongodb.server("missing_server")
    with pytest.raises(ValueError):
        client = missing_server._get_connection()
        assert not client.database_names



"""A simple command-line tool for shleem."""

import pprint

import click

from shleem.param_inference_maps import rebuild_all_maps
from shleem.collection_cfg import rebuild_collection_cfg_files
from shleem.shared import _shleem_cfg


@click.group(help="Utility shleem operations.")
def util():
    """Utility shleem operations."""
    pass


@util.command(help="Rebuild parameter inference maps.")
def rebuildmaps():
    """Rebuild parameter inference maps."""
    rebuild_all_maps()


@util.command(help="Rebuild collection attributes.")
def rebuildattr():
    """Rebuild collection attributes."""
    rebuild_collection_cfg_files()


@util.command(help="Print shleem's current configuration.")
def printcfg():
    """Print shleem's current configuration."""
    pprint.pprint(_shleem_cfg(), indent=1, width=10)



"""Core functions for the shleem package."""

import abc


DEFAULT_SOURCE_TYPE = 'unspecified'


class DataSource(object):
    """A base class for shleem data sources.

    Arguments
    ---------
    identifier : str
        An string identifier unique to this data source.
    source_type : str, optional
        A string identifying the type of this data source. If none is given, a
        default value is used.
    """

    def __init__(self, identifier, source_type=None):
        self.identifier = identifier
        if source_type is None:
            source_type = DEFAULT_SOURCE_TYPE
        self.source_type = source_type

    def __repr__(self):
        return "DataSource: {}".format(self.identifier)


class DataTap(DataSource, metaclass=abc.ABCMeta):
    """An abstract base class for tappable shleem data sources.

    Arguments
    ---------
    identifier : str
        An string identifier unique to this data source.
    source_type : str, optional
        A string identifying the type of this data source. If none is given, a
        default value is used.
    """

    @abc.abstractmethod
    def tap(self, **kwargs):
        """Taps this DataTap to produce a raw dataset."""
        pass  # pragma: no cover

    def __repr__(self):
        return "DataTap: {}".format(self.identifier)



"""Shared functionalities for the shleem package."""

import os


HOMEDIR = os.path.expanduser("~")
SHLEEM_DIR_NAME = '.shleem'
SHLEEM_DIR_PATH = os.path.join(HOMEDIR, SHLEEM_DIR_NAME)
os.makedirs(SHLEEM_DIR_PATH, exist_ok=True)

SHLEEM_CFG_FNAME = 'config.json'
SHLEEM_CFG_FPATH = os.path.join(SHLEEM_DIR_PATH, SHLEEM_CFG_FNAME)



"""Testing core functionalities of the shleem package."""

from shleem import (
    DataSource,
    DataTap,
)


def test_data_source():
    some_id = "232"
    some_source = DataSource(identifier=some_id)
    assert some_source.identifier == some_id
    assert some_source.source_type == 'unspecified'
    assert some_source.__repr__() == 'DataSource: {}'.format(some_id)

    some_type = "ShleemDB"
    source2 = DataSource(identifier=some_id, source_type=some_type)
    assert source2.identifier == some_id
    assert source2.source_type == some_type


def test_data_tap():
    class ShleemDBQuery(DataTap):
        def __init__(self, query):
            super().__init__(
                identifier="ShleemDB."+str(query), source_type="ShleemDB")
            self.query = query
        def tap(self):
            return []
    shuq = ShleemDBQuery({'a': 4})
    assert shuq.source_type == "ShleemDB"
    assert shuq.__repr__() == "DataTap: ShleemDB.{'a': 4}"
    assert shuq.tap() == []



"""Automate and version datasets generation from data sources."""
# pylint: disable=C0413,C0411

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

# === module imports

from .core import ( # noqa
    DataSource,
    DataTap,
)

import shleem.mongodb  # noqa: E402, F401

for name in ['shleem', 'core', 'shared']:
    try:
        globals().pop(name)
    except KeyError:
        pass
try:
    del name  # pylint: disable=W0631
except NameError:
    pass



"""MongoDB shleem data sources."""

import sys

from .mongodb import (
    server,
    _add_servers_attr,
)

_add_servers_attr(sys.modules[__name__])

for name in [
        'sys', 'mongodb', '_add_servers_attr',
]:
    try:
        globals().pop(name)
    except KeyError:
        pass
try:
    del name  # pylint: disable=W0631
except NameError:
    pass



"""Exceptions used by the shleem package."""


class ConfigurationException(Exception):  # pragma: no cover
    """An exception thrown when a problem is found in a configuration mapping
    used to configure shleem."""
    pass



"""A command-line interface for shleem."""

import click

from .util_cli import util


@click.group(help="A command-line interface for shleem.")
def cli():
    """A command-line interface for shleem."""
    pass


cli.add_command(util)



"""MongoDB data sources."""

import os
import copy
import json
import urllib.parse
from functools import lru_cache


from pymongo import MongoClient
from strct.general import stable_hash_builtins_strct

from shleem.core import (
    DataSource,
    DataTap,
)
from shleem.shared import SHLEEM_DIR_PATH


MONGODB_SOURCE_TYPE = 'MongoDB'
SHLEEM_MONGODB_CRED_FNAME = 'mongodb_credentials.json'
SHLEEM_MONGODB_CRED_FPATH = os.path.join(
    SHLEEM_DIR_PATH, SHLEEM_MONGODB_CRED_FNAME)
MONGO_CRED_FILE_MSG = (
    'MongoDB credentials for shleem should be set by a credentials file. The '
    'credentials file should be named mongodb_credentials.json, placed in the '
    '.shleem folder located in your home folder and constructed in the '
    'following format:\n'
    '{\n'
    '    "servers": {\n'
    '        "production_data_store": {\n'
    '            "hosts": ["ds192763.mlab.com:51829"],\n'
    '             "username": "shleem_reader",\n'
    '             "password": "8d728673tfi8h723yds"\n'
    '        }\n'
    '    }\n'
    '}'
)


def _get_cred():
    """Returns MongoDB credentials dict."""
    try:
        with open(SHLEEM_MONGODB_CRED_FPATH, 'r') as mongo_cred_file:
            return json.load(mongo_cred_file)
    except FileNotFoundError:  # pragma: no cover
        print(MONGO_CRED_FILE_MSG)


class MongoDBSource(DataSource):
    """A MongoDB data source.

    Arguments
    ---------
    identifier : str
        A string identifier unique to this data source.
    """

    def __init__(self, identifier):
        DataSource.__init__(
            self, identifier=identifier, source_type=MONGODB_SOURCE_TYPE)


class MongoDBServer(MongoDBSource):
    """A MongoDB server data source.

    Arguments
    ---------
    server_name : str
        The name of this MongoDB server.
    """

    def __init__(self, server_name):
        MongoDBSource.__init__(self, identifier=server_name)
        self.server_name = server_name

    def __repr__(self):
        return "MongoDB server DataSource: {}".format(self.identifier)

    def __getitem__(self, db_name):
        return self.db(db_name)

    def __getattr__(self, db_name):
        return self.db(db_name)

    @lru_cache(maxsize=1024)
    def db(self, db_name):
        """Returns a MongoDBDataBase object with the given name, hosted on this
        MongoDB server."""
        return MongoDBDatabase(self, db_name)

    @staticmethod
    def _mongodb_uris(usr, pwd, hosts):
        parsed_usr = urllib.parse.quote_plus(usr)
        parsed_pwd = urllib.parse.quote_plus(pwd)
        return [
            "mongodb://{usr}:{pwd}@{host}".format(
                usr=parsed_usr, pwd=parsed_pwd, host=host)
            for host in hosts
        ]

    @lru_cache(maxsize=2)
    def _get_connection(self):
        """Returns a pymongo client connected to this server.

        Returns
        -------
        pymongo.MongoClient
            Returns a pymongo.MongoClient object with reading permissions
            connected to this server.
        """
        cred = copy.deepcopy(_get_cred())
        try:
            server_cred = cred['servers'][self.server_name]
            uris = MongoDBServer._mongodb_uris(
                usr=server_cred.pop('username'),
                pwd=server_cred.pop('password'),
                hosts=server_cred.pop('hosts'),
            )
            return MongoClient(host=uris, **server_cred)
        except KeyError:
            msg = ("The server {} is missing for shleem's MongoDB credentials"
                   "file.\n".format(self.server_name) + MONGO_CRED_FILE_MSG)
            raise ValueError(msg)


@lru_cache(maxsize=1024)
def server(server_name):
    """Returns a MongoDBServer object with the given name."""
    return MongoDBServer(server_name)


def _add_servers_attr(module):
    cred = _get_cred()
    for server_name in cred['servers']:
        setattr(module, server_name, server(server_name))


class MongoDBDatabase(MongoDBSource):
    """A specific MongoDB database data source.

    Arguments
    ---------
    mongodb_server : MongoDBServer
        The MongoDB server this database is located on.
    db_name : str
        The name of the database.
    """

    def __init__(self, mongodb_server, db_name):
        identifier = mongodb_server.identifier + '.' + db_name
        MongoDBSource.__init__(self, identifier=identifier)
        self.mongodb_server = mongodb_server
        self.db_name = db_name

    def __repr__(self):
        return "MongoDB database DataSource: {}".format(self.identifier)

    @lru_cache(maxsize=1024)
    def __getitem__(self, collection_name):
        return MongoDBCollection(self, collection_name)

    def __getattr__(self, collection_name):
        return self[collection_name]

    def collection(self, collection_name):
        """Returns a MongoDBCollection data source object with the given name,
        located at this MongoDB database."""
        return self[collection_name]

    @lru_cache(maxsize=2)
    def _get_connection(self):
        """Returns a pymongo.database.Database object connected to this
        database."""
        return self.mongodb_server._get_connection()[self.db_name]


class MongoDBCollection(MongoDBSource):
    """A specific MongoDB collection data source.

    Arguments
    ---------
    mongodb_db : MongoDBDatabase
        The MongoDB database this collection is located on.
    collection_name : str
        The name of the collection.
    """

    def __init__(self, mongodb_db, collection_name):
        identifier = mongodb_db.identifier + '.' + collection_name
        MongoDBSource.__init__(self, identifier=identifier)
        self.mongodb_db = mongodb_db
        self.collection_name = collection_name

    def __repr__(self):
        return "MongoDB collection DataSource: {}".format(self.identifier)

    def query(self, query_dict, identifier=None, projection=None, skip=None,
              limit=None):
        """Returns a MongoDBQuery source object representing a query ran
        against this collection.

        Arguments
        ---------
        query_dict : dict
            A pymongo-compliant MongoDB query.
        identifier : str, optional
            A string identifier unique to this query. If none is given, a
            stable hash function is used to compute a good candidate.
        projection : list or dict, optional
            A list of field names that should be returned in the result set or
            a dict specifying the fields to include or exclude.
        skip : int, optional
            the number of documents to omit (from the start of the result set)
            when returning the results.
        limit : int, optional
            the maximum number of results to return.
        """
        return MongoDBQuery(self, query=query_dict, identifier=identifier,
                            projection=projection, skip=skip, limit=limit)

    def aggregation(self, aggregation_pipeline, identifier=None):
        """Returns a MongoDBAggregation source object representing an
        aggregation ran against this collection.

        Arguments
        ---------
        aggregation_piepline : list
            A pymongo-compliant MongoDB aggregation given as a list of dicts.
        identifier : str, optional
            A string identifier unique to this aggregation. If none is given, a
            stable hash function is used to compute a good candidate.
        """
        return MongoDBAggregation(
            self, aggregation_pipeline=aggregation_pipeline,
            identifier=identifier)

    @lru_cache(maxsize=2)
    def _get_connection(self):
        """Returns a pymongo.collection.Collection object connected to this
        database."""
        return self.mongodb_db._get_connection()[self.collection_name]


def _clean_query_helper(obj):
    if isinstance(obj, dict):
        for key in obj:
            if callable(obj[key]):
                obj[key] = obj[key].__name__
            elif isinstance(obj[key], (dict, list, tuple)):
                obj[key] = _clean_query_helper(obj[key])
    if isinstance(obj, (list, tuple)):
        new_list = []
        for item in obj:
            if callable(item):
                new_list.append(item.__name__)
            elif isinstance(item, (dict, list, tuple)):
                new_list.append(_clean_query_helper(item))
            else:
                new_list.append(item)
        return new_list
    return obj


def _clean_query_from_callables(query):
    new = copy.deepcopy(query)
    return _clean_query_helper(new)


def _resolve_helper(obj, **kwargs):
    if isinstance(obj, dict):
        for key in obj.keys():
            if callable(obj[key]):
                obj[key] = obj[key](**kwargs)
            if isinstance(obj[key], (dict, list, tuple)):
                obj[key] = _resolve_helper(obj[key], **kwargs)
        return obj
    # else, it's a list or a tuple
    # if isinstance(obj, (list, tuple)):
    new_list = []
    for item in obj:
        if callable(item):
            new_list.append(item(**kwargs))
        elif isinstance(item, (dict, list, tuple)):
            new_list.append(_resolve_helper(item, **kwargs))
        else:
            new_list.append(item)
    return new_list


def _resolve_query(query, **kwargs):
    resolved_query = copy.deepcopy(query)
    return _resolve_helper(resolved_query, **kwargs)


class MongoDBQuery(MongoDBSource, DataTap):
    """A specific MongoDB query data source.

    Objects of this class should not be instantiated directly, but rather using
    the query method of shleem.MongoDBCollection objects.

    Arguments
    ---------
    mongodb_collection : MongoDBCollection
        The MongoDB collection this query will be ran against.
    query : dict
        A pymongo-compliant MongoDB query.
    identifier : str, optional
        A string identifier unique to this query. If none is given, a stable
        hash function is used to compute a good candidate.
    projection : list or dict, optional
        A list of field names that should be returned in the result set or a
        dict specifying the fields to include or exclude.
    skip : int, optional
        the number of documents to omit (from the start of the result set)
        when returning the results.
    limit : int, optional
        the maximum number of results to return.
    """

    def __init__(self, mongodb_collection, query, identifier=None,
                 projection=None, skip=None, limit=None):
        if identifier is None:
            try:
                identifier = str(stable_hash_builtins_strct(query))
            except TypeError:
                clean_query = _clean_query_from_callables(query)
                identifier = str(stable_hash_builtins_strct(clean_query))
        identifier = mongodb_collection.identifier + '.' + identifier
        super().__init__(identifier=identifier)
        self.mongodb_collection = mongodb_collection
        self.query = query
        self.projection = projection
        if skip is None:
            skip = 0
        if limit is None:
            limit = 0  # pargma: no cover
        self.skip = skip
        self.limit = limit

    def __repr__(self):
        return "MongoDB query DataSource: {}".format(self.identifier)

    def tap(self, **kwargs):
        col_obj = self.mongodb_collection._get_connection()
        query = _resolve_query(self.query, **kwargs)
        return col_obj.find(
            filter=query,
            projection=self.projection,
            skip=self.skip,
            limit=self.limit,
        )


class MongoDBAggregation(MongoDBSource, DataTap):
    """A specific MongoDB aggregation data source.

    Arguments
    ---------
    mongodb_collection : MongoDBCollection
        The MongoDB collection this query will be ran against.
    aggregation_pipeline : list
        A pymongo-compliant MongoDB aggregation pipelien, given as a list of
        dicts.
    identifier : str, optional
        A string identifier unique to this aggregation. If none is given, a
        stable hash function is used to compute a good candidate.
    """

    def __init__(self, mongodb_collection, aggregation_pipeline,
                 identifier=None):
        if identifier is None:
            try:
                identifier = str(stable_hash_builtins_strct(
                    aggregation_pipeline))
            except TypeError:
                clean_pipe = _clean_query_from_callables(
                    aggregation_pipeline)
                identifier = str(stable_hash_builtins_strct(clean_pipe))
        identifier = mongodb_collection.identifier + '.' + identifier
        super().__init__(identifier=identifier)
        self.mongodb_collection = mongodb_collection
        self.aggregation_pipeline = aggregation_pipeline

    def __repr__(self):
        return "MongoDB aggregation DataSource: {}".format(self.identifier)

    def tap(self, **kwargs):
        col_obj = self.mongodb_collection._get_connection()
        pipe = _resolve_query(self.aggregation_pipeline, **kwargs)
        return col_obj.aggregate(pipe)
